/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <stdio.h>
#include <sstream>
#include <digitalSTROM/ds485-client-interface.h>

#include "ds-assistant.h"
#include "exceptions.h"
#include "database.h"
#include "util.h"
#include "dsmapi-wrapper.h"
#include "html_report_web_handler.h"
#include "gettext_helper.h"
#include "storage.h"
#include "tasks.h"

#define MAX_FETCH_RECORDS                   50

namespace dSA
{

HTMLReportWebHandler::HTMLReportWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               HTMLWebHandler(connection, request_info) {}


std::string HTMLReportWebHandler::taskOverview(
                                        boost::shared_ptr<dSLongTaskInfo> task)
{
    std::ostringstream result;

    result << header(_("Overview"));

    result << table_begin();

    result << table_row_begin();
    result << table_column_add(task->getName(), 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add(task->getDescription() + "<br/>", 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add(_("Status:"), 0, 250);
    result << table_column_add(task->getStatusInfo());
    result << table_row_end();

    task_status t = task->getTaskStatus();
    if (t == Queued)
    {
        result << table_end();
        return result.str();
    }

    result << table_row_simple(_("Started:"),
        timeutil::localizedEpoch2local((time_t)task->getStartedTime()));

    if ((t != Cancelled) && (t != Running))
    {
        result << table_row_simple(_("Ended:"),
            timeutil::localizedEpoch2local((time_t)task->getFinishedTime()));
    }

    if (t == Running)
    {
        result << table_row_begin();
        result << table_column_add(_("<b>Note:</b> a protocol can only be "
                                   "generated after the test has completed."),
                                   2);
        result << table_row_end();

        result << table_end();
        return result.str();
    }


    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::upstreamPercent(int up)
{
    return util::str(up * 100 / TRANSMISSION_QUALITY_UPSTREAM_MAX) + '%';
}

std::string HTMLReportWebHandler::downstreamPercent(int down)
{
    return util::str(100 - (down * 100 / TRANSMISSION_QUALITY_DOWNSTREAM_MAX)) +
           '%';
}

std::string HTMLReportWebHandler::transmissionQualitySummary(long long int id)
{
    std::ostringstream result;

    boost::shared_ptr<Database::task_result_summary> summary =
        Database::getInstance()->getTaskResultSummary(id);

    boost::shared_ptr<Database::result_data> data = summary->data;

    result << header(_("Summary"));
    result << table_begin();

    result << table_row_begin();
    result << table_column_add(_("Number of tested devices:"), 0, 250);
    result << table_column_add((*data)["devices"]->first);
    result << table_row_end();

    result << table_row_simple(_("Number of executed tests:"),
                               (*data)["tests"]->first);

    result << table_row_simple(_("Number of tests with errors:"),
                               (*data)["errors"]->first);

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();


    result << table_row_simple(_("Average upstream quality:"),
                upstreamPercent(util::str2int((*data)["avgUpstream"]->first)));

    result << table_row_simple(_("Maximum upstream quality:"),
                upstreamPercent(util::str2int((*data)["maxUpstream"]->first)));

    result << table_row_simple(_("Minimum upstream quality:"),
                upstreamPercent(util::str2int((*data)["minUpstream"]->first)));

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();

    result << table_row_simple(_("Average downstream quality:"),
            downstreamPercent(util::str2int((*data)["avgDownstream"]->first)));

    result << table_row_simple(_("Maximum downstream quality:"),
            downstreamPercent(util::str2int((*data)["maxDownstream"]->first)));

    result << table_row_simple(_("Minimum downstream quality:"),
            downstreamPercent(util::str2int((*data)["minDownstream"]->first)));

    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::renderTQDevREcord(
        boost::shared_ptr<
            const Database::transmission_quality_device_record> rec)
{
    std::ostringstream result;

    result << table_row_begin();
    try
    {
        Storage *storage = Storage::getInstance();
        boost::shared_ptr<Device> device =
                            storage->getDevice(util::str2dsuid(rec->dsuid));
        result << table_column_add(device->getDisplayId());
    }
    catch (Exception &ex)
    {
        result << table_column_add(util::render_display_id(
                                                util::str2dsuid(rec->dsuid)));
    }

    if (rec->errorcode == 0)
    {
        result << table_column_add(upstreamPercent(rec->upstream));
        result << table_column_add(downstreamPercent(rec->downstream));
    }
    else
    {
        result << table_column_add("-");
        result << table_column_add("-");
    }
    result << table_column_add(ds485c_strerror(rec->errorcode));
    result << table_column_add(timeutil::localizedEpoch2local(rec->timestamp));
    result << table_row_end();
    return result.str();
}

void HTMLReportWebHandler::transmissionQualityReport(
                                        boost::shared_ptr<dSLongTaskInfo> task,
                                            bool download)
{
    long long int c = 0;
    std::string s = page_begin(_("Protocol of transmission quality"), download);
    mg_write(m_connection, s.c_str(), s.length());

    try
    {
        s = taskOverview(task);
        mg_write(m_connection, s.c_str(), s.length());

        task_status t = task->getTaskStatus();
        if ((t != Finished) && (t != Cancelled))
        {
            s = page_end();
            mg_write(m_connection, s.c_str(), s.length());
            return;
        }

        s = transmissionQualitySummary(task->getTaskID());
        mg_write(m_connection, s.c_str(), s.length());

        Database* db = Database::getInstance();
        c = db->transmissionQualityResultCount(task->getTaskID());
        long long int offset = 0;

        std::ostringstream html;

        if (c > 0)
        {
            s = header(_("Detailed results"));
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();

            html << table_row_begin();
            html << table_header(_("Device ID"));
            html << table_header(_("Upstream"));
            html << table_header(_("Downstream"));
            html << table_header(_("Result"));
            html << table_header(_("Time"));
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::transmission_quality_device_record> > >
                        records = db->getTransmissionQualityResults(
                            task->getTaskID(), MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::transmission_quality_device_record> rec =
                        records->front();
                records->pop_front();
                html << renderTQDevREcord(rec);
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception &ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error(_("Could not complete report:") + std::string(" ") +
                  ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::testDevicePresenceReport(
        boost::shared_ptr<dSLongTaskInfo> task,
        bool download)
{
    long long int c = 0;
    std::string s = page_begin(_("Protocol of device presence"), download);
    mg_write(m_connection, s.c_str(), s.length());
    Storage *storage = Storage::getInstance();

    try
    {
        s = taskOverview(task);
        mg_write(m_connection, s.c_str(), s.length());

        task_status t = task->getTaskStatus();
        if ((t != Finished) && (t != Cancelled))
        {
            s = page_end();
            mg_write(m_connection, s.c_str(), s.length());
            return;
        }

        Database* db = Database::getInstance();
        boost::shared_ptr<TestDevicePresenceTask> dptask = boost::dynamic_pointer_cast<TestDevicePresenceTask> (task);
        webutil::query params;
        webutil::parse_query(dptask->getParameters(), params);
        dsuid_t dsmdsuid = util::str2dsuid(params["dsm"]);

        // Summary
        boost::shared_ptr<Database::task_result_summary> summary =
                db->getTaskResultSummary(task->getTaskID());
        {
            std::ostringstream result;
            boost::shared_ptr<Database::result_data> data = summary->data;

            result << header(_("Summary"));
            result << table_begin();

            if (!dsuid_is_null(&dsmdsuid))
            {
                result << table_row_begin();
                result << table_column_add(_("dSM ID of the power circuit:"),
                                           0, 250);
                try
                {
                    Storage *storage = Storage::getInstance();
                    boost::shared_ptr<dSM> dsm = storage->getDSM(dsmdsuid);
                    result << table_column_add(dsm->getDisplayId());
                }
                catch (Exception &ex)
                {
                    result << table_column_add(util::render_display_id(dsmdsuid));
                }
                result << table_row_end();
            }
            result << table_row_begin();
            result << table_column_add(_("Number of tested devices:"), 0, 250);
            result << table_column_add((*data)["tests"]->first);
            result << table_row_end();

            c = util::str2llint((*data)["tests"]->first);
            long long int replies = util::str2llint((*data)["replies"]->first);
            long long int active = util::str2llint((*data)["active"]->first);

            result << table_row_simple(_("Number of newly detected devices:"),
                    (*data)["active"]->first);
            result << table_row_simple(_("Number of removed devices:"),
                    (*data)["inactive"]->first);
            result << table_row_simple(
                                _("Number of active devices after the test:"),
                                util::str(replies + active));

            result << table_row_begin();
            result << table_column_add("&nbsp;", 2);
            result << table_row_end();

            result << table_end();

            s = result.str();
            mg_write(m_connection, s.c_str(), s.length());
        }

        // Details
        long long int offset = 0;
        std::ostringstream html;

        if (c > 0)
        {
            s = header(_("Detailed results"));
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();
            html << table_row_begin();
            if (dsuid_is_null(&dsmdsuid))
            {
                html << table_header(_("dSM ID"));
            }
            html << table_header(_("Device ID"));
            html << table_header(_("Result"));
            html << table_header(_("Time"));
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::device_presence_record> > >
                        records = db->getDevicePresenceResults(
                            task->getTaskID(), MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::device_presence_record> rec =
                        records->front();
                records->pop_front();

                std::ostringstream result;
                result << table_row_begin();
                if (dsuid_is_null(&dsmdsuid))
                {
                    try
                    {
                        boost::shared_ptr<dSM> dsm =
                            storage->getDSM(util::str2dsuid(rec->dsmdsuid));
                        result << table_column_add(dsm->getDisplayId());
                    }
                    catch (Exception &ex)
                    {
                        result << table_column_add(
                                util::render_display_id(
                                    util::str2dsuid(rec->dsmdsuid)));
                    }
                }

                try
                {
                    boost::shared_ptr<Device> device =
                            storage->getDevice(util::str2dsuid(rec->dsuid));
                    result << table_column_add(device->getDisplayId());
                }
                catch (Exception &ex)
                {

                    result << table_column_add(
                            util::render_display_id(
                                util::str2dsuid(rec->dsuid)));
                }

                switch(rec->resultcode)
                {
                    case 5:
                        result << table_column_add(
                                std::string(_("Inactive (was active before)")));
                        break;
                    case 6:
                        result << table_column_add(
                                std::string(_("Active (was inactive before)")));
                        break;
                    case 4:
                        result << table_column_add(std::string(_("Active")));
                        break;
                    case 3:
                        result << table_column_add(std::string(_("Inactive")));
                        break;
                    case 2:
                        result << table_column_add(
                                std::string(_("Not queried")));
                        break;
                }
                result << table_column_add(
                        timeutil::localizedEpoch2local(rec->timestamp));
                result << table_row_end();
                html << result.str();
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception &ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error(_("Could not complete report: ") + ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::testDeviceUpdateReport(
        boost::shared_ptr<dSLongTaskInfo> task,
        bool download)
{
    long long int c = 0;
    std::string s = page_begin(_("Status of device configuration update"), download);
    mg_write(m_connection, s.c_str(), s.length());
    Storage *storage = Storage::getInstance();
    try
    {
        {
            std::ostringstream result;
            result << header(_("Overview"));
            result << table_begin();

            result << table_row_begin();
            result << table_column_add(task->getName());
            result << table_column_add(task->getDescription());
            result << table_row_end();

            result << table_row_begin();
            result << table_column_add(_("Progress:"));
            result << table_column_add(task->getProgressInfo());
            result << table_row_end();

            if (task->isPaused())
            {
                result << table_row_begin();
                result << table_column_add(_("Status:"));
                result << table_column_add(_("Paused"));
                result << table_row_end();
            }

            result << table_end();
            s = result.str();
        }

        mg_write(m_connection, s.c_str(), s.length());

        Database* db = Database::getInstance();
        boost::shared_ptr<DeviceUpdateTask> dptask = boost::dynamic_pointer_cast<DeviceUpdateTask> (task);

        if (task->getProgress() < 100)
        {
            std::ostringstream result;
            std::string scList;
            std::vector<int> scenes = dptask->getSceneChanges();

            result << header(_("Configuration Update"));
            result << table_begin();

            result << table_row_begin();
            result << table_column_add(_("Total number of scenes to update:"));
            result << table_column_add(util::str(scenes.size()));
            result << table_row_end();

            for (std::vector<int>::iterator it = scenes.begin(); it != scenes.end(); it ++)
            {
                scList += " " + util::str(*it) + " ";
            }

            result << table_row_begin();
            result << table_column_add(_("Scenes in progress:"));
            result << table_column_add(scList);
            result << table_row_end();

            result << table_end();
            s = result.str();
            mg_write(m_connection, s.c_str(), s.length());
        }

        webutil::query params;
        webutil::parse_query(dptask->getParameters(), params);

        // Summary
        boost::shared_ptr<Database::task_result_summary> summary =
                db->getTaskResultSummary(task->getTaskID());
        {
            std::ostringstream result;
            boost::shared_ptr<Database::result_data> data = summary->data;

            result << header(_("Summary"));
            result << table_begin();

            result << table_row_begin();
            result << table_column_add(_("Total number of devices:"));
            result << table_column_add((*data)["devcount"]->first);
            result << table_row_end();

            result << table_row_simple(_("Number of devices with updated configuration:"),
                    (*data)["uptodate"]->first);
            result << table_row_simple(_("Number of devices that cannot be updated:"),
                    (*data)["incompatible"]->first);
            result << table_row_simple(_("Number of inactive devices:"),
                    (*data)["inactive"]->first);
            result << table_row_simple(_("Number of devices with communication errors:"),
                    (*data)["commerror"]->first);
            result << table_end();

            s = result.str();
            mg_write(m_connection, s.c_str(), s.length());

            c = util::str2llint((*data)["devcount"]->first);
        }

        // Details
        long long int offset = 0;
        std::ostringstream html;

        if (c > 0)
        {
            s = header(_("Detailed results"));
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();
            html << table_row_begin();

            html << table_header(_("Device ID"));
            html << table_header(_("Meter ID"));
            html << table_header(_("Device Name"));
            html << table_header(_("Firmware Revision"));
            html << table_header(_("Configuration ID"));
            html << table_header(_("Status"));
            html << table_header(_("Last Status Change"));
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::device_configupdate_record> > >
                        records = db->getDeviceConfigUpdateResults(
                            /*task->getTaskID(),*/ MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::device_configupdate_record> rec =
                        records->front();
                records->pop_front();

                std::ostringstream result;
                result << table_row_begin();
                try
                {
                    boost::shared_ptr<Device> device =
                            storage->getDevice(util::str2dsuid(rec->dsuid));
                    result << table_column_add(device->getDisplayId());
                }
                catch (Exception &ex)
                {
                    result << table_column_add(util::render_display_id(
                                               util::str2dsuid(rec->dsuid)));
                }

                try
                {
                    boost::shared_ptr<dSM> dsm =
                            storage->getDSM(util::str2dsuid(rec->dsmdsuid));
                    result << table_column_add(dsm->getDisplayId());

                }
                catch (Exception &ex)
                {
                    result << table_column_add(util::render_display_id(
                                               util::str2dsuid(rec->dsmdsuid)));
                }
                result << table_column_add(rec->devname);
                result << table_column_add(util::hex(rec->firmware, 4));
                switch (rec->configid)
                {
                    case 255:
                        result << table_column_add(std::string(_("-")));
                        break;
                    default:
                        result << table_column_add(util::str(rec->configid));
                        break;
                }
                switch (rec->status)
                {
                    case 1:
                        result << table_column_add(std::string(_("Unknown")));
                        break;
                    case 3:
                        result << table_column_add(std::string(_("In Progress")));
                        break;
                    case 2:
                    case 4:
                        result << table_column_add(std::string(_("Up to date")));
                        break;
                    case 5:
                    {
                        switch (rec->errorcode)
                        {
                            case 1:
                                result << table_column_add(std::string(_("No Error")));
                                break;
                            case 2:
                                result << table_column_add(std::string(_("Inactive")));
                                break;
                            case 3:
                                result << table_column_add(std::string(_("Communication Error")));
                                break;
                            default:
                                result << table_column_add(std::string(_("Unknown Error")));
                                break;
                        }
                    }
                    break;
                }
                result << table_column_add(
                        timeutil::localizedEpoch2local(rec->timestamp));
                result << table_row_end();
                html << result.str();
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception &ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error(_("Could not complete report: ") + ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::taskReport()
{
    std::string task_id = m_parameters["id"];
    std::string task_typ = m_parameters["type"];
    if (task_id.empty() && task_typ.empty())
    {
        failure(_("Error"),
                _("Can not generate report: missing task id or type parameter!"));
        return;
    }

    bool dl = false;
    std::string download = m_parameters["download"];
    if (!download.empty())
    {
        try
        {
            dl = (util::str2int(download));
        }
        catch (Exception &ex)
        {
            failure(_("Error"),
                    _("Can not generate report: invalid download parameter!"));
            return;
        }
    }

    boost::shared_ptr<dSLongTaskInfo> task;
    try
    {
        if (task_id.length())
        {
            long long int tid = util::str2llint(task_id);
            task = dS_Assistant::getInstance()->getTaskInfo(tid);
            if (NULL == task)
            {
                task = Database::getInstance()->getTask(tid);
            }
        }
        else if (task_typ.length())
        {
            task_type ttyp = (task_type) util::str2llint(task_typ);
            task = dS_Assistant::getInstance()->getTaskInfoByType(ttyp);
            if (NULL == task)
            {
                boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > > tasklist =
                        Database::getInstance()->getTasksByType(ttyp);
                if (!tasklist->empty())
                {
                    task = tasklist->front();
                }
            }
        }
    }
    catch (Exception &ex)
    {
        failure(_("Error"), _("Can not generate report!") + std::string(" ") +
                ex.getMessage());
        return;
    }

    if (NULL == task)
    {
        failure(_("Error"),
                _("Can not generate report: cannot find task!"));
        return;
    }

    switch (task->getTaskType())
    {
        case TestTransmissionQuality:
            transmissionQualityReport(task, dl);
            break;
        case TestDevicePresence:
            testDevicePresenceReport(task, dl);
            break;
        case TestDeviceUpdateTask:
            testDeviceUpdateReport(task, dl);
            break;
        default:
            failure(_("Error"),
                    _("Can not generate report: unknown task type!"));
            return;
    }
}

void HTMLReportWebHandler::process()
{
    // unlike the json calls the HTML report can probably get quite big,
    // therefore we will try to keep a lesser amount of data in memory and
    // avoid passing large strings around, instead each report has to
    // write chunks of data using mg_write()
    try
    {
        if (m_command == "taskReport")
        {
            taskReport();
        }
        else
        {
            failure(_("Error - can not generate report!"),
                    _("Unknown function called"));
        }
    }
    // we can't return anything to the user at this point because data is
    // written out in chunks, meaning that the user received at least a part of
    // the page already; the functions that take care of processing each call
    // have to ensure that the html never gets broken and thus handle all
    // possible exceptions gracefully
    catch (Exception &ex)
    {
        fprintf(stderr, "Error generating report: %s\n",
                ex.getMessage().c_str());
    }
    catch (...)
    {
        fprintf(stderr, "Encountered unexpected error while processing call");
    }
}

} // namespace

