/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Michael Tross <mt@aizo.ag>, based on http://www.drdobbs.com/cpp/201804215

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __LOG_H__
#define __LOG_H__

#include <sstream>
#include <string>
#include <stdio.h>
#include <sys/time.h>
#include <boost/thread/mutex.hpp>
#include "util.h"

namespace dSA
{

inline std::string NowTime(time_t t0 = 0)
{
    char buffer[32];
    time_t t;
    if (t0 == 0) {
        time(&t);
    } else {
        t = t0;
    }
    struct tm r;
    memset(&r, 0, sizeof(r));
    strftime(buffer, sizeof(buffer), "%F %T", localtime_r(&t, &r));
    struct timeval tv;
    gettimeofday(&tv, 0);
    char result[100];
    sprintf(result, "%s.%03ld", buffer, (long) tv.tv_usec / 1000);
    return result;
}

enum TLogLevel
{
    logERROR,
    logWARNING,
    logINFO,
    logDEBUG,
    logDEBUG1,
    logDEBUG2,
    logDEBUG3,
    logDEBUG4
};


template<typename T>
class Log
{
public:
    Log();
    virtual ~Log();
    std::ostringstream& Get(TLogLevel level = logINFO);
    std::ostringstream& GetDSUID(const dsuid_t dsuid,
                                TLogLevel level = logINFO,
                                time_t t0 = 0);
    std::ostringstream& GetDSUIDPrefix(const dsuid_t dsuid, const char *prefix,
                                TLogLevel level = logINFO);
    std::ostringstream& GetDSUIDData(const dsuid_t dsuid, const char *prefix);
public:
    static TLogLevel& ReportingLevel();
    static std::string ToString(TLogLevel level);
    static TLogLevel FromString(const std::string& level);
protected:
    std::ostringstream os;
private:
    Log(const Log&);
    Log& operator =(const Log&);
};

template<typename T>
Log<T>::Log()
{
}

template<typename T>
std::ostringstream& Log<T>::Get(TLogLevel level)
{
    os << "[" << NowTime(0) << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSUID(const dsuid_t dsuid, TLogLevel level, time_t t0)
{
    os << "[" << NowTime(t0) << "]";
    ///os << "[" << util::str(dsuid) << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    T::SetDSUID(util::str(dsuid));
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSUIDPrefix(const dsuid_t dsuid, const char *prefix, TLogLevel level)
{
    os << "[" << NowTime(0) << "]";
    ///os << "[" << util::str(dsuid) << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    T::SetDSUIDPrefix(util::str(dsuid), prefix);
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSUIDData(const dsuid_t dsuid, const char *prefix)
{
    os << NowTime(0) << ",";
    T::SetDSUIDPrefix(util::str(dsuid), prefix);
    return os;
}

template<typename T>
Log<T>::~Log()
{
    os << std::endl;
    T::Output(os.str());
}

template<typename T>
TLogLevel& Log<T>::ReportingLevel()
{
    static TLogLevel reportingLevel = logDEBUG4;
    return reportingLevel;
}

template<typename T>
std::string Log<T>::ToString(TLogLevel level)
{
    static const char* const buffer[] =
    {
        "ERROR", "WARNING", "INFO", "DEBUG", "DEBUG1", "DEBUG2", "DEBUG3",
        "DEBUG4"
    };
    return buffer[level];
}

template<typename T>
TLogLevel Log<T>::FromString(const std::string& level)
{
    if (level == "DEBUG4")
        return logDEBUG4;
    if (level == "DEBUG3")
        return logDEBUG3;
    if (level == "DEBUG2")
        return logDEBUG2;
    if (level == "DEBUG1")
        return logDEBUG1;
    if (level == "DEBUG")
        return logDEBUG;
    if (level == "INFO")
        return logINFO;
    if (level == "WARNING")
        return logWARNING;
    if (level == "ERROR")
        return logERROR;
    Log<T> ().Get(logWARNING) << "Unknown logging level '" << level
            << "'. Using INFO level as default.";
    return logINFO;
}

class Output2FILE
{
public:
    static void Output(const std::string& msg);
private:
    static FILE*& StreamImpl();
    static boost::mutex& MutexImpl();
};

inline FILE*& Output2FILE::StreamImpl()
{
    static FILE* pStream = stderr;
    return pStream;
}

inline boost::mutex& Output2FILE::MutexImpl()
{
    static boost::mutex m_mutex;
    return m_mutex;
}

inline void Output2FILE::Output(const std::string& msg)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    FILE* pStream = StreamImpl();
    if (!pStream)
        return;
    fprintf(pStream, "%s", msg.c_str());
    fflush(pStream);
}

typedef Log<Output2FILE> FILELog;

class Output2DSUID
{
public:
    static void Output(const std::string& msg);
    static void SetDSUID(const std::string dsuid);
    static void SetDSUIDPrefix(const std::string dsuid, const char *prefix);
    static void SetLogdirectory(const std::string dir);
private:
    static FILE*& StreamImpl();
    static boost::mutex& MutexImpl();
    static std::string& LogdirImpl();
};

inline FILE*& Output2DSUID::StreamImpl()
{
    static FILE* pStream = stderr;
    return pStream;
}

inline boost::mutex& Output2DSUID::MutexImpl()
{
    static boost::mutex m_mutex;
    return m_mutex;
}

inline std::string& Output2DSUID::LogdirImpl()
{
    static std::string m_logdir;
    return m_logdir;
}

inline void Output2DSUID::SetDSUID(const std::string dsuid)
{
    MutexImpl().lock();
    if (LogdirImpl().empty())
    {
        StreamImpl() = stderr;
        return;
    }
    std::string filename(LogdirImpl() + "/dsuid_" +
                         dsuid.c_str() + std::string(".log"));
    StreamImpl() = fopen(filename.c_str(), "a");
}

inline void Output2DSUID::SetDSUIDPrefix(const std::string dsuid, const char *prefix)
{
    MutexImpl().lock();
    if (LogdirImpl().empty())
    {
        StreamImpl() = stderr;
        return;
    }
    std::string filename(LogdirImpl() + "/" + std::string(prefix) +
                         dsuid.c_str() + std::string(".log"));
    StreamImpl() = fopen(filename.c_str(), "a");
}

inline void Output2DSUID::SetLogdirectory(const std::string dir)
{
    LogdirImpl() = dir;
}

inline void Output2DSUID::Output(const std::string& msg)
{
    FILE* pStream = StreamImpl();
    if (!pStream)
    {
        MutexImpl().unlock();
        return;
    }
    fprintf(pStream, "%s", msg.c_str());
    fflush(pStream);
    if ((pStream != stderr) && (pStream != NULL))
    {
        fclose(pStream);
        StreamImpl() = NULL;
    }
    MutexImpl().unlock();
}

typedef Log<Output2DSUID> DSUIDLog;

#ifndef FILELOG_MAX_LEVEL
#ifdef NDEBUG
# define FILELOG_MAX_LEVEL logDEBUG
#else
# define FILELOG_MAX_LEVEL logDEBUG4
#endif
#endif

#define FILE_LOG(level) \
    if (level > FILELOG_MAX_LEVEL) ;\
    else if (level > FILELog::ReportingLevel()) ; \
    else FILELog().Get(level)

#ifndef DSUID_MAX_LEVEL
#ifdef NDEBUG
# define DSUID_MAX_LEVEL logDEBUG
#else
# define DSUID_MAX_LEVEL logDEBUG4
#endif
#endif

#define DSUID_LOG(level, dsuid) \
    if (level > DSUID_MAX_LEVEL) ;\
    else if (level > DSUIDLog::ReportingLevel()) ; \
    else DSUIDLog().GetDSUID(dsuid, level, 0)

#define DSUID_DATA_LOG(dsuid, _prefix) \
    DSUIDLog().GetDSUIDData(dsuid, _prefix)

#define DSUID_LOG_TS(level, dsuid, timestamp) \
    if (level > DSUID_MAX_LEVEL) ;\
    else if (level > DSUIDLog::ReportingLevel()) ; \
    else DSUIDLog().GetDSUID(dsuid, level, timestamp)

#define DSUID_PREFIX_LOG(_level, _dsuid, _prefix) \
    if (_level > DSUID_MAX_LEVEL) ;\
    else if (_level > DSUIDLog::ReportingLevel()) ; \
    else DSUIDLog().GetDSUIDPrefix(_dsuid, _prefix, _level)

};

#endif //__LOG_H__
